# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_trainer.ipynb (unless otherwise specified).

__all__ = ['MODEL_DICT', 'OPT_DICT', 'create_classifier', 'create_cnn', 'Trainer', 'InterpretModel', 'Learner']

# Cell
from typing import Callable, Union

import tensorflow as tf
import tensorflow_addons as tfa
from tensorflow.keras.models import Model
from typeguard import check_argument_types, check_return_type, typechecked

from .datagenerator import Dataset

# Cell
from functools import partial

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from tf_keras_vis.gradcam import Gradcam, GradcamPlusPlus
from tf_keras_vis.utils import normalize

# Cell
import inspect
import sys

MODEL_DICT = {}
for name, func in inspect.getmembers(tf.keras.applications):
    if inspect.isfunction(func):
        MODEL_DICT[name.lower()] = func

OPT_DICT = {}
for name, func in inspect.getmembers(tf.keras.optimizers):
    if inspect.isclass(func):
        OPT_DICT[name.lower()] = func

# Cell
@typechecked
def _get_base_cnn(
    base_model: Union[str, Model],
    pooling: str = "avg",
    weights: Union[str, None] = "imagenet",
    include_top: bool = False,
) -> Model:
    if isinstance(base_model, str):
        assert (
            base_model in MODEL_DICT.keys()
        ), f"base_model name must be in {tuple(MODEL_DICT.keys())}"
        base_model = MODEL_DICT[base_model]
        base_model = base_model(
            include_top=include_top, pooling=pooling, weights=weights
        )
    return base_model

# Cell
@typechecked
def _add_output_layers(
    base_model: Model, outputs: int, drop_out: Union[float, None] = None, name=None
) -> Model:

    x = base_model.output
    # x = tf.keras.layers.GlobalMaxPool2D()(x)
    if drop_out:
        x = tf.keras.layers.Dropout(drop_out)(x)
    x = tf.keras.layers.Dense(outputs, name="output")(x)

    model = tf.keras.Model(base_model.input, x, name=name)
    return model

# Cell
def create_classifier(
    base_model_fn: callable,
    num_classes: int,
    weights="imagenet",
    dropout=0,
    include_top=False,
    name=None,
):

    outputs = 1 if num_classes == 2 else num_classes

    base_model = base_model_fn(
        include_top=include_top,
        weights=weights,
    )
    if include_top:
        return base_model
    drop_out = 0.5
    outputs = 1

    x = base_model.output
    x = tf.keras.layers.GlobalMaxPool2D()(x)
    if drop_out:
        x = tf.keras.layers.Dropout(drop_out)(x)
    x = tf.keras.layers.Dense(outputs, name="output")(x)

    model = tf.keras.Model(base_model.input, x)
    return model

# Cell
@typechecked
def create_cnn(
    base_model: Union[str, Model],
    num_classes: int,
    drop_out=0.5,
    keras_applications: bool = True,
    pooling: str = "avg",
    weights: Union[str, None] = "imagenet",
    name=None,
) -> Model:

    assert pooling in ("avg", "max")

    if keras_applications:
        if num_classes == 2:
            outputs = 1
        else:
            outputs = num_classes
    else:
        print(f"num_classes is ignored. returning the passed model as it is.")

    if isinstance(base_model, (str, Model)) and keras_applications:
        base_model = _get_base_cnn(base_model, pooling=pooling, weights=weights)
        assert (
            "pool" in base_model.layers[-1].name
        ), f"base_model last layer must be a pooling layer"
        model = _add_output_layers(base_model, outputs, drop_out=drop_out, name=name)

    elif isinstance(base_model, Model) and keras_applications is False:
        model = base_model

    elif isinstance(base_model, str) and keras_applications is False:
        model = _get_base_cnn(base_model, weights="imagenet", include_top=True)

    else:
        print(f"Invalid arguments!")
    return model

# Cell
class Trainer(Model):
    """
    The Trainer class inherits tf.keras.Model and contains everything a model needs for training.
    It exposes trainer.cyclic_fit method which trains the model using Cyclic Learning rate discovered by Leslie Smith.

    Arguments:
    ds: Dataset object
    model: object of type tf.keras.Model
    num_classes (int, None): number of classes in the dataset. If None then will auto infer from Dataset

    """

    _AUTOTUNE = tf.data.experimental.AUTOTUNE

    @typechecked
    def __init__(
        self, ds: Dataset, model: Model, num_classes: Union[int, None] = None, **kwargs
    ):
        assert check_argument_types()

        super(Trainer, self).__init__()
        self.ds = ds
        self.total = len(ds)
        if num_classes:
            self.NUM_CLASSES = num_classes
        else:
            self.NUM_CLASSES = ds.NUM_CLASSES
        self.gradcam = None
        self.model = model
        self.cyclic_opt_set = False

    def build(self):
        pass

    def summary(self):
        return self.model.summary()

    # def get_layer(name=None, index=None): return self.model(name, index)

    def compile(self, *args, **kwargs):
        return self.model.compile(*args, **kwargs)

    def call(self, *args, **kwargs):
        return self.model.call(*args, **kwargs)

    def fit(self, *args, **kwargs):
        return self.model.fit(*args, **kwargs)

    def warmup(self):
        pass

    def prewhiten(self, image):
        image = tf.cast(image, tf.float32)
        image = image / 127.5 - 1.0
        return image

    def rescale(self, image, label):
        image = self.prewhiten(image)
        return image, label

    def _get_optimizer(self, optimizer, momentum=0.9, **kwargs):
        if optimizer.__name__ == "SGD":
            optimizer = partial(
                optimizer, momentum=momentum, nesterov=kwargs.get("nesterov", True)
            )
        else:
            optimizer = partial(
                optimizer,
                momentum=momentum,
            )
        return optimizer

    def _prepare_dl(self, bs=8, shuffle=True):
        ds = self.ds
        dl = ds.get_tf_dataset(shuffle=shuffle)
        dl = dl.map(self.rescale, Trainer._AUTOTUNE)
        return dl.batch(bs).prefetch(Trainer._AUTOTUNE)

    def cyclic_fit(
        self,
        epochs: int,
        batch_size: int,
        lr_range: Union[tuple, list] = (1e-4, 1e-2),
        optimizer=tf.keras.optimizers.SGD,
        momentum=0.9,
        validation_data=None,
        callbacks=None,
        *args,
        **kwargs,
    ):
        """Trains model on ds as train data with cyclic learning rate.
        Dataset will be automatically converted into `tf.data` format and images will be prewhitened in range of [-1, 1].
        Cyclical Learning Rates for Training Neural Networks: https://arxiv.org/abs/1506.01186

        Args:
            epochs (int): number of epochs for training
            batch_size (int): batch size
            lr_range (tuple): learning rate will cycle from lr_min to lr_max
            optimizer (callable): Keras callable optimizer
            momentum(int): momentum for the optimizer
        kwargs:
            step_size (int): step size for the Cyclic learning rate. By default it is `2 * len(self.ds)//batch_size`
            scale_mode (str): cycle or exp
            shuffle(bool): Dataset will be shuffle on each epoch if True
        """
        if not self.cyclic_opt_set:
            self.max_lr, self.min_lr = lr_range
            ds = self.ds
            step_size = 2 * len(self.ds) // batch_size
            lr_schedule = tfa.optimizers.Triangular2CyclicalLearningRate(
                initial_learning_rate=lr_range[0],
                maximal_learning_rate=lr_range[1],
                step_size=kwargs.get("step_size", step_size),
                scale_mode=kwargs.get("scale_mode", "cycle"),
            )

            optimizer = self._get_optimizer(optimizer, momentum=momentum)
            optimizer = optimizer(learning_rate=lr_schedule)
            self.model.optimizer = optimizer
            self.cyclic_opt_set = True
        else:
            print("cyclic learning rate already set!")

        return self.model.fit(
            self._prepare_dl(batch_size, kwargs.get("shuffle", True)),
            validation_data=validation_data,
            epochs=epochs,
            callbacks=callbacks,
        )

    @typechecked
    def compile2(
        self,
        batch_size: int,
        optimizer: Union[str, tf.keras.optimizers.Optimizer] = "adam",
        lr_range: Union[tuple, list] = (1e-4, 1e-2),
        loss=None,
        metrics=None,
        loss_weights=None,
        weighted_metrics=None,
        run_eagerly=None,
        **kwargs,
    ):
        """Compile2 compiles the model of Trainer for cyclic learning rate.
        Cyclical Learning Rates for Training Neural Networks: https://arxiv.org/abs/1506.01186

        Args:
            batch_size (int): batch size
            lr_range (tuple): learning rate will cycle from lr_min to lr_max
            optimizer (str, keras.optimizer.Optimizer): Keras optimizer

        kwargs:
            step_size (int): step size for the Cyclic learning rate. By default it is `2 * len(self.ds)//batch_size`
            scale_mode (str): cycle or exp
            momentum(int): momentum for the optimizer when optimizer is of type str
        """
        self.max_lr, self.min_lr = lr_range
        self.batch_size = batch_size

        self.step_size = step_size = 2 * len(self.ds) // batch_size

        lr_schedule = tfa.optimizers.Triangular2CyclicalLearningRate(
            initial_learning_rate=lr_range[0],
            maximal_learning_rate=lr_range[1],
            step_size=kwargs.get("step_size", step_size),
            scale_mode=kwargs.get("scale_mode", "cycle"),
        )

        if isinstance(optimizer, str):
            optimizer = OPT_DICT[optimizer]
            optimizer = optimizer(learning_rate=lr_schedule)
            if kwargs.get("momentum"):
                optimizer.momentum = kwargs.get("momentum")

        else:
            optimizer.learning_rate = lr_schedule

        self.compile(optimizer=optimizer, loss=loss, metrics=metrics)
        self.cyclic_opt_set = True
        print(f"Model compiled!")

# Cell
class InterpretModel:
    def __init__(self, gradcam_pp: bool, learner: Trainer, clone: bool = False):
        """Args:
        gradcam_pp: if True GradCam class will be used else GradCamPlusplus
        clone: whether GradCam will clone learner.model
        """
        if gradcam_pp:
            self.gradcam_fn = GradcamPlusPlus
        else:
            self.gradcam_fn = Gradcam
        self.learner = learner

        self.gradcam = self.gradcam_fn(learner.model, self.model_modifier, clone=clone)

        # if self.learner.include_top is not True:
        #     self.gradcam._find_penultimate_output = self.patch

    def __call__(self, image: Image.Image, auto_resize: bool = True, image_size=None):
        # assert check_argument_types()
        gradcam = self.gradcam
        get_loss = self.get_loss
        if auto_resize and image_size is None:
            image_size = self.learner.ds.img_sz_list.get_size()
        if image_size:
            image = image.resize(image_size)

        X = np.asarray(image, np.float32)
        X = self.learner.prewhiten(X)
        X = np.expand_dims(X, 0)

        cam = gradcam(
            get_loss,
            X,
            penultimate_layer=-1,  # model.layers number
            seek_penultimate_conv_layer=True,
        )
        cam = normalize(cam)
        heatmap = np.uint8(cm.jet(cam[0])[..., :3] * 255)
        plt.imshow(image)
        plt.imshow(heatmap, cmap="jet", alpha=0.5)
        plt.show()

    def __patch(self, *args, **kwargs):
        """Path _find_penultimate_output method of tf_keras_vis"""
        if self.learner.include_top:
            return self.learner.model.layers[-1].output
        return self.learner.model.layers[0].get_output_at(-1)

    def model_modifier(self, m):
        """Sets last activation to linear"""
        m.layers[-1].activation = tf.keras.activations.linear
        return m

    def get_loss(self, preds):
        if self.learner.NUM_CLASSES == 2:
            ret = preds[0]
        else:
            index = tf.argmax(tf.math.softmax(preds), axis=1)[0]
            # print(index, preds.shape)
            ret = preds[0, index]
            print(f"index: {index}")
        return ret

# Cell
from typing import List, Union

from tensorflow import keras
import pytorch_lightning as pl
from .utility.import_utils import INSTALLED_MODULES
from .converter.core import tf2_to_onnx, tf2_to_pytorch, onnx_to_pytorch, pytorch_to_onnx

# Cell
class Learner:
    TF = ("TF", "TENSORFLOW")
    PT = ("PYTORCH", "PT", "TORCH")

    def __init__(self,
                 model: Union[pl.LightningModule, keras.models.Model],
                 mode: str = "TF"):
        self.MODE = mode.upper()
        self.model = model
        self.epochs_trained = 0

        if self.MODE in Learner.PT:
            self.trainer = None

    def fit(self,
            train_data,
            epochs,
            val_data=None,
            test_data=None,
            callbacks=None,
            **kwargs):
        """train models
        For TF:
            Just pass train data and start training
        For PyTorch:
            You can enter configs to Lightning Trainer
        """
        MODE = self.MODE
        initial_epoch = self.epochs_trained
        self.epochs_trained += epochs

        if MODE in Learner.TF:
            return self.model.fit(
                train_data,
                epochs=epochs,
                initial_epoch=initial_epoch,
                validation_data=val_data,
                callbacks=callbacks,
            )
        elif MODE in Learner.PT:
            lit_confs = kwargs.get('LIT_TRAINER_CONFIG', {})
            if not self.trainer:
                self.trainer = pl.Trainer(max_epochs=epochs, **lit_confs)
            return self.trainer.fit(self.model, train_data, val_data)

    def to_onnx(self, tensor=None, export_path=None):
        MODE = self.MODE
        if MODE in Learner.TF:
            return tf2_to_onnx(self.model, output_path=export_path)

        if MODE in Learner.PT:
            return pytorch_to_onnx(self.model, tensor, export_path)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_tf_utils.ipynb (unless otherwise specified).

__all__ = ['disable_gpu', 'limit_gpu', 'gpu_dynamic_mem_growth']

# Cell
import tensorflow as tf
import os

# Cell
def disable_gpu():
    """
    disable gpu for tensorflow
    """
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = str(-1)


#export
def limit_gpu(gpu_id: str, memory_limit:int):
    """
    limit the selected gpu [gpu_id] by [memory_limit] MB
    """
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_id)
    gpus = tf.config.list_physical_devices('GPU')

    if gpus:
        # Restrict TensorFlow to only allocate [memory MB] of memory on the first GPU
        try:
            tf.config.experimental.set_virtual_device_configuration(
                gpus[0], [
                    tf.config.experimental.VirtualDeviceConfiguration(
                        memory_limit=memory_limit)
                ])
            logical_gpus = tf.config.list_logical_devices('GPU')
            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Virtual devices must be set before GPUs have been initialized
            print(e)
    else:
        print(f'No GPU:{gpu_id} found in your system!')



# Cell
def gpu_dynamic_mem_growth():
    """
    Borrowed from https://github.com/philipperemy/keract/tree/master/examples
    """
    # Check for GPUs and set them to dynamically grow memory as needed
    # Avoids OOM from tensorflow greedily allocating GPU memory
    try:
        gpu_devices = tf.config.list_physical_devices('GPU')
        if len(gpu_devices) > 0:
            for gpu in gpu_devices:
                tf.config.experimental.set_memory_growth(gpu, True)
            print('GPU dynamic memory growth enabled')
        else: print('No GPU found on the machine!')
    except AttributeError:
        print('Upgrade your tensorflow to 2.x to have the gpu_dynamic_mem_growth feature.')


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"remove_dsstore": "00_core.ipynb",
         "get_basename": "04_dataloader.ipynb",
         "IMAGENET_LABELS": "00_core.ipynb",
         "tf": "01_image.ipynb",
         "torch": "01_image.ipynb",
         "DATA_FORMATS": "01_image.ipynb",
         "DEFAULT_MODE": "01_image.ipynb",
         "read_image": "01_image.ipynb",
         "resize_image": "01_image.ipynb",
         "BoundingBoxes": "01_image.ipynb",
         "Chitra": "01_image.ipynb",
         "benchmark": "02_datagenerator.ipynb",
         "get_filenames": "02_datagenerator.ipynb",
         "get_label": "02_datagenerator.ipynb",
         "ImageSizeList": "02_datagenerator.ipynb",
         "Pipeline": "02_datagenerator.ipynb",
         "Dataset": "02_datagenerator.ipynb",
         "MODEL_DICT": "03_trainer.ipynb",
         "OPT_DICT": "03_trainer.ipynb",
         "create_classifier": "03_trainer.ipynb",
         "create_cnn": "03_trainer.ipynb",
         "Trainer": "03_trainer.ipynb",
         "InterpretModel": "03_trainer.ipynb",
         "Learner": "03_trainer.ipynb",
         "AUTOTUNE": "04_dataloader.ipynb",
         "show_batch": "04_dataloader.ipynb",
         "Clf": "04_dataloader.ipynb",
         "label_color": "05_visualization.ipynb",
         "colors": "05_visualization.ipynb",
         "draw_box": "05_visualization.ipynb",
         "draw_caption": "05_visualization.ipynb",
         "draw_boxes": "05_visualization.ipynb",
         "draw_detections": "05_visualization.ipynb",
         "draw_annotations": "05_visualization.ipynb",
         "disable_gpu": "06_tf_utils.ipynb",
         "limit_gpu": "06_tf_utils.ipynb",
         "gpu_dynamic_mem_growth": "06_tf_utils.ipynb",
         "is_installed": "07_import_utils.ipynb",
         "INSTALLED_MODULES": "07_import_utils.ipynb",
         "pytorch_to_onnx": "08_model_converter.ipynb",
         "onnx_to_pytorch": "08_model_converter.ipynb",
         "tf2_to_onnx": "08_model_converter.ipynb",
         "tf2_to_pytorch": "08_model_converter.ipynb"}

modules = ["core.py",
           "image.py",
           "datagenerator.py",
           "trainer.py",
           "dataloader.py",
           "visualization.py",
           "utility/tf_utils.py",
           "utility/import_utils.py",
           "converter/core.py"]

doc_url = "https://chitra.aniketmaurya.com/"

git_url = "https://github.com/aniketmaurya/chitra/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_import_utils.ipynb (unless otherwise specified).

__all__ = ['is_installed', 'INSTALLED_MODULES']

# Cell
import importlib

# Cell
def is_installed(module_name: str):
    return importlib.util.find_spec(module_name) is not None


_TF = "tensorflow"
_TF_GPU = "tensorflow-gpu"
_TORCH = "torch"
_TORCHVISION = "torchvision"
_JAX = "jax"
_FLAX = "flax"


INSTALLED_MODULES = {
    module: is_installed(module)
    for module in (_TF, _TF_GPU, _TORCH, _TORCHVISION, _JAX, _FLAX)
}


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_image.ipynb (unless otherwise specified).

__all__ = ['tf', 'torch', 'DATA_FORMATS', 'DEFAULT_MODE', 'read_image', 'resize_image', 'BoundingBoxes', 'Chitra']

# Cell
import os
from io import BytesIO
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import requests
from PIL import Image

from .utility.import_utils import INSTALLED_MODULES, _TF, _TORCH

tf = None
torch = None

if INSTALLED_MODULES.get(_TF, None):
    import tensorflow as tf

if INSTALLED_MODULES.get(_TORCH, None):
    import torch

# Cell
from typing import List, Union, Optional

DATA_FORMATS = Union[str, Image.Image, np.ndarray, tf.Tensor, torch.Tensor]
DEFAULT_MODE = os.environ.get("CHITRA_DEFAULT_MODE", "TF")

from imgaug.augmentables.bbs import BoundingBox, BoundingBoxesOnImage

# Cell
def read_image(path: str, channels: int = 3):
    """Reads an image file from the path and return the rgb image in tf.Tensor format."""
    img: tf.Tensor = tf.io.read_file(path)
    img: tf.Tensor = tf.io.decode_image(img,
                                        channels=channels,
                                        expand_animations=False)
    return img


def resize_image(image: tf.Tensor, size: Union[tf.Tensor, tuple], **kwargs):
    """Resize image to the target `size`: Union[tf.Tensor, tuple]"""
    assert isinstance(
        image,
        tf.Tensor), f"image must be of type tf.Tensor but passed {type(image)}"
    assert isinstance(size, (tuple, tf.Tensor))
    method = kwargs.get("method", "bilinear")
    return tf.image.resize(image, size, method)


def _url_to_image(url: str) -> Image.Image:
    """returns Image from url"""
    assert url.lower().startswith("http"), "invalid url, must start with http"
    content = requests.get(url).content
    image = Image.open(BytesIO(content))
    return image

# Cell
class BoundingBoxes:

    CENTER = "XXYY"
    CORNER = "XYXY"

    def __init__(self,
                 bboxes: Optional[List[list]] = None,
                 labels: Optional[List[Union[int, str]]] = None,
                 format: str = 'xyxy'):
        """Args:
            bboxes: list of bounding boxes [(x1, y1, x2, y2), ...] or [(xc, yc, h, w), ...]
            labels: list of strings or integers
            format:
                - `xyxy` for corner points of bbox
                - `xyhw` for x-center, y-center, height and width format of bbox
        """
        assert format.upper() in (
            self.CENTER,
            self.CORNER), f"bbox format must be either xyxy or xyhw"
        bboxes = self._listify(bboxes, 4)
        labels = self._listify(labels)
        assert len(bboxes) == len(
            labels
        ), f"len of boxes and labels not matching: {len(bboxes), len(labels)}"

        self._format = format.upper()
        self.bboxes = self._list_to_bbox(bboxes, labels)
        self._state = {}

    def _listify(self, item, dim_trigger=None):
        if item is None:
            return item

        if not isinstance(item, (list, tuple)):
            return [item]

        if isinstance(item, (list, tuple)):
            if self.num_dim(item) == dim_trigger:
                item = [item]
        return item

    @staticmethod
    def num_dim(item):
        return len(item)

    @staticmethod
    def center_to_corner(cx, cy, h, w):
        xmin = cx - w / 2
        xmax = cx + w / 2
        ymin = cy - h / 2
        ymax = cy + h / 2

        return xmin, ymin, xmax, ymax

    @staticmethod
    def corner_to_center(xmin, ymin, xmax, ymax):
        w = xmax - xmin
        h = ymax - ymin

        cx = xmin + w / 2
        cy = ymin + h / 2

        return cx, cy, h, w

    def _list_to_bbox(
            self,
            bbox_list: Optional[List[List[Union[int, float]]]],
            labels: List[Union[str, int]] = None) -> List[BoundingBox]:
        """Converts bbox list into `imgaug BoundigBox` object
        """
        format = self._format

        if not bbox_list:
            return None

        if not labels:
            labels = [None] * self.num_dim(bbox_list)

        bbox_objects = []
        for bbox, label in zip(bbox_list, labels):
            if format == self.CENTER:
                bbox = self.center_to_corner(*bbox)
            bbox_objects.append(BoundingBox(*bbox, label))
        return bbox_objects

    def __getitem__(self, idx):
        return self.bboxes[idx]

    def __repr__(self):
        return str(self.bboxes)

    def get_bounding_boxes_on_image(self, image_shape):
        """returns `imgaug BoundingBoxesOnImage` object which can be used to boxes on the image
        """
        return BoundingBoxesOnImage(self.bboxes, image_shape)

# Cell
class Chitra:
    """Ultimate image utility class.
          1. Load image from file, web url, numpy or bytes
          2. Plot image
          3. Draw bounding boxes
    """

    def __init__(self, data, bboxes=None, labels=None, format=BoundingBoxes.CORNER, *args, **kwargs) -> None:
        """Args:
            data: numpy, url, filelike
        """
        super().__init__()
        self.image = self._load_image(data)
        self.bboxes = None

        if bboxes is not None:
            self.bboxes = BoundingBoxes(bboxes, labels)

    def _load_image(self, data: DATA_FORMATS):
        if isinstance(data, Image.Image):
            return data

        if isinstance(data, (tf.Tensor, torch.Tensor)):
            data = data.numpy()

        if isinstance(data, str):
            if data.startswith("http"):
                image = _url_to_image(data)
            else:
                image = Image.open(data)

        elif isinstance(data, np.ndarray):
            image = Image.fromarray(data)

        else:
            raise UserWarning("unable to load image!")

        return image

    def numpy(self):
        return np.asarray(self.image)

    def to_tensor(self, mode: str = DEFAULT_MODE):
        """mode: tf/torch/pt"""
        mode = mode.upper()
        np_image = self.numpy()
        tensor = None

        if mode == "TF":
            tensor = tf.constant(np_image)
        elif mode in ("TORCH", "PT"):
            tensor = torch.from_numpy(np_image)
        else:
            raise UserWarning("invalid mode!")
        return tensor

    @property
    def shape(self):
        return self.numpy().shape

    @property
    def size(self):
        return self.image.size

    def imshow(self, cmap=plt.cm.Blues, *args, **kwargs):
        plt.imshow(self.numpy(), cmap, *args, **kwargs)

    def draw_boxes(
            self,
            marker_size: int = 2,
            color=(0, 255, 0),
    ):
        if self.bboxes is None:
            raise UserWarning('bboxes is None')

        bbox_on_image = self.bboxes.get_bounding_boxes_on_image(self.shape)
        return bbox_on_image.draw_on_image(self.numpy(),
                                           color=color,
                                           size=marker_size)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = ['remove_dsstore', 'get_basename', 'IMAGENET_LABELS']

# Cell
import tensorflow as tf
import numpy as np
from glob import glob
import os
import pathlib

from typing import Union

# Cell
def remove_dsstore(path):
    """
    Deletes .DS_Store files from path and sub-folders of path.
    """
    path = pathlib.Path(path)

    for e in path.glob('*.DS_Store'):
        os.remove(e)

    for e in path.glob('*/*.DS_Store'):
        os.remove(e)


def get_basename(path: tf.string):
    assert isinstance(path, tf.Tensor)
    return tf.strings.split(path, os.path.sep)[-1]

# Cell
IMAGENET_LABELS = ["tench",
            "goldfish",
            "great white shark",
            "tiger shark",
            "hammerhead shark",
            "electric ray",
            "stingray",
            "cock",
            "hen",
            "ostrich",
            "brambling",
            "goldfinch",
            "house finch",
            "junco",
            "indigo bunting",
            "American robin",
            "bulbul",
            "jay",
            "magpie",
            "chickadee",
            "American dipper",
            "kite",
            "bald eagle",
            "vulture",
            "great grey owl",
            "fire salamander",
            "smooth newt",
            "newt",
            "spotted salamander",
            "axolotl",
            "American bullfrog",
            "tree frog",
            "tailed frog",
            "loggerhead sea turtle",
            "leatherback sea turtle",
            "mud turtle",
            "terrapin",
            "box turtle",
            "banded gecko",
            "green iguana",
            "Carolina anole",
            "desert grassland whiptail lizard",
            "agama",
            "frilled-necked lizard",
            "alligator lizard",
            "Gila monster",
            "European green lizard",
            "chameleon",
            "Komodo dragon",
            "Nile crocodile",
            "American alligator",
            "triceratops",
            "worm snake",
            "ring-necked snake",
            "eastern hog-nosed snake",
            "smooth green snake",
            "kingsnake",
            "garter snake",
            "water snake",
            "vine snake",
            "night snake",
            "boa constrictor",
            "African rock python",
            "Indian cobra",
            "green mamba",
            "sea snake",
            "Saharan horned viper",
            "eastern diamondback rattlesnake",
            "sidewinder",
            "trilobite",
            "harvestman",
            "scorpion",
            "yellow garden spider",
            "barn spider",
            "European garden spider",
            "southern black widow",
            "tarantula",
            "wolf spider",
            "tick",
            "centipede",
            "black grouse",
            "ptarmigan",
            "ruffed grouse",
            "prairie grouse",
            "peacock",
            "quail",
            "partridge",
            "grey parrot",
            "macaw",
            "sulphur-crested cockatoo",
            "lorikeet",
            "coucal",
            "bee eater",
            "hornbill",
            "hummingbird",
            "jacamar",
            "toucan",
            "duck",
            "red-breasted merganser",
            "goose",
            "black swan",
            "tusker",
            "echidna",
            "platypus",
            "wallaby",
            "koala",
            "wombat",
            "jellyfish",
            "sea anemone",
            "brain coral",
            "flatworm",
            "nematode",
            "conch",
            "snail",
            "slug",
            "sea slug",
            "chiton",
            "chambered nautilus",
            "Dungeness crab",
            "rock crab",
            "fiddler crab",
            "red king crab",
            "American lobster",
            "spiny lobster",
            "crayfish",
            "hermit crab",
            "isopod",
            "white stork",
            "black stork",
            "spoonbill",
            "flamingo",
            "little blue heron",
            "great egret",
            "bittern",
            "crane (bird)",
            "limpkin",
            "common gallinule",
            "American coot",
            "bustard",
            "ruddy turnstone",
            "dunlin",
            "common redshank",
            "dowitcher",
            "oystercatcher",
            "pelican",
            "king penguin",
            "albatross",
            "grey whale",
            "killer whale",
            "dugong",
            "sea lion",
            "Chihuahua",
            "Japanese Chin",
            "Maltese",
            "Pekingese",
            "Shih Tzu",
            "King Charles Spaniel",
            "Papillon",
            "toy terrier",
            "Rhodesian Ridgeback",
            "Afghan Hound",
            "Basset Hound",
            "Beagle",
            "Bloodhound",
            "Bluetick Coonhound",
            "Black and Tan Coonhound",
            "Treeing Walker Coonhound",
            "English foxhound",
            "Redbone Coonhound",
            "borzoi",
            "Irish Wolfhound",
            "Italian Greyhound",
            "Whippet",
            "Ibizan Hound",
            "Norwegian Elkhound",
            "Otterhound",
            "Saluki",
            "Scottish Deerhound",
            "Weimaraner",
            "Staffordshire Bull Terrier",
            "American Staffordshire Terrier",
            "Bedlington Terrier",
            "Border Terrier",
            "Kerry Blue Terrier",
            "Irish Terrier",
            "Norfolk Terrier",
            "Norwich Terrier",
            "Yorkshire Terrier",
            "Wire Fox Terrier",
            "Lakeland Terrier",
            "Sealyham Terrier",
            "Airedale Terrier",
            "Cairn Terrier",
            "Australian Terrier",
            "Dandie Dinmont Terrier",
            "Boston Terrier",
            "Miniature Schnauzer",
            "Giant Schnauzer",
            "Standard Schnauzer",
            "Scottish Terrier",
            "Tibetan Terrier",
            "Australian Silky Terrier",
            "Soft-coated Wheaten Terrier",
            "West Highland White Terrier",
            "Lhasa Apso",
            "Flat-Coated Retriever",
            "Curly-coated Retriever",
            "Golden Retriever",
            "Labrador Retriever",
            "Chesapeake Bay Retriever",
            "German Shorthaired Pointer",
            "Vizsla",
            "English Setter",
            "Irish Setter",
            "Gordon Setter",
            "Brittany",
            "Clumber Spaniel",
            "English Springer Spaniel",
            "Welsh Springer Spaniel",
            "Cocker Spaniels",
            "Sussex Spaniel",
            "Irish Water Spaniel",
            "Kuvasz",
            "Schipperke",
            "Groenendael",
            "Malinois",
            "Briard",
            "Australian Kelpie",
            "Komondor",
            "Old English Sheepdog",
            "Shetland Sheepdog",
            "collie",
            "Border Collie",
            "Bouvier des Flandres",
            "Rottweiler",
            "German Shepherd Dog",
            "Dobermann",
            "Miniature Pinscher",
            "Greater Swiss Mountain Dog",
            "Bernese Mountain Dog",
            "Appenzeller Sennenhund",
            "Entlebucher Sennenhund",
            "Boxer",
            "Bullmastiff",
            "Tibetan Mastiff",
            "French Bulldog",
            "Great Dane",
            "St. Bernard",
            "husky",
            "Alaskan Malamute",
            "Siberian Husky",
            "Dalmatian",
            "Affenpinscher",
            "Basenji",
            "pug",
            "Leonberger",
            "Newfoundland",
            "Pyrenean Mountain Dog",
            "Samoyed",
            "Pomeranian",
            "Chow Chow",
            "Keeshond",
            "Griffon Bruxellois",
            "Pembroke Welsh Corgi",
            "Cardigan Welsh Corgi",
            "Toy Poodle",
            "Miniature Poodle",
            "Standard Poodle",
            "Mexican hairless dog",
            "grey wolf",
            "Alaskan tundra wolf",
            "red wolf",
            "coyote",
            "dingo",
            "dhole",
            "African wild dog",
            "hyena",
            "red fox",
            "kit fox",
            "Arctic fox",
            "grey fox",
            "tabby cat",
            "tiger cat",
            "Persian cat",
            "Siamese cat",
            "Egyptian Mau",
            "cougar",
            "lynx",
            "leopard",
            "snow leopard",
            "jaguar",
            "lion",
            "tiger",
            "cheetah",
            "brown bear",
            "American black bear",
            "polar bear",
            "sloth bear",
            "mongoose",
            "meerkat",
            "tiger beetle",
            "ladybug",
            "ground beetle",
            "longhorn beetle",
            "leaf beetle",
            "dung beetle",
            "rhinoceros beetle",
            "weevil",
            "fly",
            "bee",
            "ant",
            "grasshopper",
            "cricket",
            "stick insect",
            "cockroach",
            "mantis",
            "cicada",
            "leafhopper",
            "lacewing",
            "dragonfly",
            "damselfly",
            "red admiral",
            "ringlet",
            "monarch butterfly",
            "small white",
            "sulphur butterfly",
            "gossamer-winged butterfly",
            "starfish",
            "sea urchin",
            "sea cucumber",
            "cottontail rabbit",
            "hare",
            "Angora rabbit",
            "hamster",
            "porcupine",
            "fox squirrel",
            "marmot",
            "beaver",
            "guinea pig",
            "common sorrel",
            "zebra",
            "pig",
            "wild boar",
            "warthog",
            "hippopotamus",
            "ox",
            "water buffalo",
            "bison",
            "ram",
            "bighorn sheep",
            "Alpine ibex",
            "hartebeest",
            "impala",
            "gazelle",
            "dromedary",
            "llama",
            "weasel",
            "mink",
            "European polecat",
            "black-footed ferret",
            "otter",
            "skunk",
            "badger",
            "armadillo",
            "three-toed sloth",
            "orangutan",
            "gorilla",
            "chimpanzee",
            "gibbon",
            "siamang",
            "guenon",
            "patas monkey",
            "baboon",
            "macaque",
            "langur",
            "black-and-white colobus",
            "proboscis monkey",
            "marmoset",
            "white-headed capuchin",
            "howler monkey",
            "titi",
            "Geoffroy's spider monkey",
            "common squirrel monkey",
            "ring-tailed lemur",
            "indri",
            "Asian elephant",
            "African bush elephant",
            "red panda",
            "giant panda",
            "snoek",
            "eel",
            "coho salmon",
            "rock beauty",
            "clownfish",
            "sturgeon",
            "garfish",
            "lionfish",
            "pufferfish",
            "abacus",
            "abaya",
            "academic gown",
            "accordion",
            "acoustic guitar",
            "aircraft carrier",
            "airliner",
            "airship",
            "altar",
            "ambulance",
            "amphibious vehicle",
            "analog clock",
            "apiary",
            "apron",
            "waste container",
            "assault rifle",
            "backpack",
            "bakery",
            "balance beam",
            "balloon",
            "ballpoint pen",
            "Band-Aid",
            "banjo",
            "baluster",
            "barbell",
            "barber chair",
            "barbershop",
            "barn",
            "barometer",
            "barrel",
            "wheelbarrow",
            "baseball",
            "basketball",
            "bassinet",
            "bassoon",
            "swimming cap",
            "bath towel",
            "bathtub",
            "station wagon",
            "lighthouse",
            "beaker",
            "military cap",
            "beer bottle",
            "beer glass",
            "bell-cot",
            "bib",
            "tandem bicycle",
            "bikini",
            "ring binder",
            "binoculars",
            "birdhouse",
            "boathouse",
            "bobsleigh",
            "bolo tie",
            "poke bonnet",
            "bookcase",
            "bookstore",
            "bottle cap",
            "bow",
            "bow tie",
            "brass",
            "bra",
            "breakwater",
            "breastplate",
            "broom",
            "bucket",
            "buckle",
            "bulletproof vest",
            "high-speed train",
            "butcher shop",
            "taxicab",
            "cauldron",
            "candle",
            "cannon",
            "canoe",
            "can opener",
            "cardigan",
            "car mirror",
            "carousel",
            "tool kit",
            "carton",
            "car wheel",
            "automated teller machine",
            "cassette",
            "cassette player",
            "castle",
            "catamaran",
            "CD player",
            "cello",
            "mobile phone",
            "chain",
            "chain-link fence",
            "chain mail",
            "chainsaw",
            "chest",
            "chiffonier",
            "chime",
            "china cabinet",
            "Christmas stocking",
            "church",
            "movie theater",
            "cleaver",
            "cliff dwelling",
            "cloak",
            "clogs",
            "cocktail shaker",
            "coffee mug",
            "coffeemaker",
            "coil",
            "combination lock",
            "computer keyboard",
            "confectionery store",
            "container ship",
            "convertible",
            "corkscrew",
            "cornet",
            "cowboy boot",
            "cowboy hat",
            "cradle",
            "crane (machine)",
            "crash helmet",
            "crate",
            "infant bed",
            "Crock Pot",
            "croquet ball",
            "crutch",
            "cuirass",
            "dam",
            "desk",
            "desktop computer",
            "rotary dial telephone",
            "diaper",
            "digital clock",
            "digital watch",
            "dining table",
            "dishcloth",
            "dishwasher",
            "disc brake",
            "dock",
            "dog sled",
            "dome",
            "doormat",
            "drilling rig",
            "drum",
            "drumstick",
            "dumbbell",
            "Dutch oven",
            "electric fan",
            "electric guitar",
            "electric locomotive",
            "entertainment center",
            "envelope",
            "espresso machine",
            "face powder",
            "feather boa",
            "filing cabinet",
            "fireboat",
            "fire engine",
            "fire screen sheet",
            "flagpole",
            "flute",
            "folding chair",
            "football helmet",
            "forklift",
            "fountain",
            "fountain pen",
            "four-poster bed",
            "freight car",
            "French horn",
            "frying pan",
            "fur coat",
            "garbage truck",
            "gas mask",
            "gas pump",
            "goblet",
            "go-kart",
            "golf ball",
            "golf cart",
            "gondola",
            "gong",
            "gown",
            "grand piano",
            "greenhouse",
            "grille",
            "grocery store",
            "guillotine",
            "barrette",
            "hair spray",
            "half-track",
            "hammer",
            "hamper",
            "hair dryer",
            "hand-held computer",
            "handkerchief",
            "hard disk drive",
            "harmonica",
            "harp",
            "harvester",
            "hatchet",
            "holster",
            "home theater",
            "honeycomb",
            "hook",
            "hoop skirt",
            "horizontal bar",
            "horse-drawn vehicle",
            "hourglass",
            "iPod",
            "clothes iron",
            "jack-o'-lantern",
            "jeans",
            "jeep",
            "T-shirt",
            "jigsaw puzzle",
            "pulled rickshaw",
            "joystick",
            "kimono",
            "knee pad",
            "knot",
            "lab coat",
            "ladle",
            "lampshade",
            "laptop computer",
            "lawn mower",
            "lens cap",
            "paper knife",
            "library",
            "lifeboat",
            "lighter",
            "limousine",
            "ocean liner",
            "lipstick",
            "slip-on shoe",
            "lotion",
            "speaker",
            "loupe",
            "sawmill",
            "magnetic compass",
            "mail bag",
            "mailbox",
            "tights",
            "tank suit",
            "manhole cover",
            "maraca",
            "marimba",
            "mask",
            "match",
            "maypole",
            "maze",
            "measuring cup",
            "medicine chest",
            "megalith",
            "microphone",
            "microwave oven",
            "military uniform",
            "milk can",
            "minibus",
            "miniskirt",
            "minivan",
            "missile",
            "mitten",
            "mixing bowl",
            "mobile home",
            "Model T",
            "modem",
            "monastery",
            "monitor",
            "moped",
            "mortar",
            "square academic cap",
            "mosque",
            "mosquito net",
            "scooter",
            "mountain bike",
            "tent",
            "computer mouse",
            "mousetrap",
            "moving van",
            "muzzle",
            "nail",
            "neck brace",
            "necklace",
            "nipple",
            "notebook computer",
            "obelisk",
            "oboe",
            "ocarina",
            "odometer",
            "oil filter",
            "organ",
            "oscilloscope",
            "overskirt",
            "bullock cart",
            "oxygen mask",
            "packet",
            "paddle",
            "paddle wheel",
            "padlock",
            "paintbrush",
            "pajamas",
            "palace",
            "pan flute",
            "paper towel",
            "parachute",
            "parallel bars",
            "park bench",
            "parking meter",
            "passenger car",
            "patio",
            "payphone",
            "pedestal",
            "pencil case",
            "pencil sharpener",
            "perfume",
            "Petri dish",
            "photocopier",
            "plectrum",
            "Pickelhaube",
            "picket fence",
            "pickup truck",
            "pier",
            "piggy bank",
            "pill bottle",
            "pillow",
            "ping-pong ball",
            "pinwheel",
            "pirate ship",
            "pitcher",
            "hand plane",
            "planetarium",
            "plastic bag",
            "plate rack",
            "plow",
            "plunger",
            "Polaroid camera",
            "pole",
            "police van",
            "poncho",
            "billiard table",
            "soda bottle",
            "pot",
            "potter's wheel",
            "power drill",
            "prayer rug",
            "printer",
            "prison",
            "projectile",
            "projector",
            "hockey puck",
            "punching bag",
            "purse",
            "quill",
            "quilt",
            "race car",
            "racket",
            "radiator",
            "radio",
            "radio telescope",
            "rain barrel",
            "recreational vehicle",
            "reel",
            "reflex camera",
            "refrigerator",
            "remote control",
            "restaurant",
            "revolver",
            "rifle",
            "rocking chair",
            "rotisserie",
            "eraser",
            "rugby ball",
            "ruler",
            "running shoe",
            "safe",
            "safety pin",
            "salt shaker",
            "sandal",
            "sarong",
            "saxophone",
            "scabbard",
            "weighing scale",
            "school bus",
            "schooner",
            "scoreboard",
            "CRT screen",
            "screw",
            "screwdriver",
            "seat belt",
            "sewing machine",
            "shield",
            "shoe store",
            "shoji",
            "shopping basket",
            "shopping cart",
            "shovel",
            "shower cap",
            "shower curtain",
            "ski",
            "ski mask",
            "sleeping bag",
            "slide rule",
            "sliding door",
            "slot machine",
            "snorkel",
            "snowmobile",
            "snowplow",
            "soap dispenser",
            "soccer ball",
            "sock",
            "solar thermal collector",
            "sombrero",
            "soup bowl",
            "space bar",
            "space heater",
            "space shuttle",
            "spatula",
            "motorboat",
            "spider web",
            "spindle",
            "sports car",
            "spotlight",
            "stage",
            "steam locomotive",
            "through arch bridge",
            "steel drum",
            "stethoscope",
            "scarf",
            "stone wall",
            "stopwatch",
            "stove",
            "strainer",
            "tram",
            "stretcher",
            "couch",
            "stupa",
            "submarine",
            "suit",
            "sundial",
            "sunglass",
            "sunglasses",
            "sunscreen",
            "suspension bridge",
            "mop",
            "sweatshirt",
            "swimsuit",
            "swing",
            "switch",
            "syringe",
            "table lamp",
            "tank",
            "tape player",
            "teapot",
            "teddy bear",
            "television",
            "tennis ball",
            "thatched roof",
            "front curtain",
            "thimble",
            "threshing machine",
            "throne",
            "tile roof",
            "toaster",
            "tobacco shop",
            "toilet seat",
            "torch",
            "totem pole",
            "tow truck",
            "toy store",
            "tractor",
            "semi-trailer truck",
            "tray",
            "trench coat",
            "tricycle",
            "trimaran",
            "tripod",
            "triumphal arch",
            "trolleybus",
            "trombone",
            "tub",
            "turnstile",
            "typewriter keyboard",
            "umbrella",
            "unicycle",
            "upright piano",
            "vacuum cleaner",
            "vase",
            "vault",
            "velvet",
            "vending machine",
            "vestment",
            "viaduct",
            "violin",
            "volleyball",
            "waffle iron",
            "wall clock",
            "wallet",
            "wardrobe",
            "military aircraft",
            "sink",
            "washing machine",
            "water bottle",
            "water jug",
            "water tower",
            "whiskey jug",
            "whistle",
            "wig",
            "window screen",
            "window shade",
            "Windsor tie",
            "wine bottle",
            "wing",
            "wok",
            "wooden spoon",
            "wool",
            "split-rail fence",
            "shipwreck",
            "yawl",
            "yurt",
            "website",
            "comic book",
            "crossword",
            "traffic sign",
            "traffic light",
            "dust jacket",
            "menu",
            "plate",
            "guacamole",
            "consomme",
            "hot pot",
            "trifle",
            "ice cream",
            "ice pop",
            "baguette",
            "bagel",
            "pretzel",
            "cheeseburger",
            "hot dog",
            "mashed potato",
            "cabbage",
            "broccoli",
            "cauliflower",
            "zucchini",
            "spaghetti squash",
            "acorn squash",
            "butternut squash",
            "cucumber",
            "artichoke",
            "bell pepper",
            "cardoon",
            "mushroom",
            "Granny Smith",
            "strawberry",
            "orange",
            "lemon",
            "fig",
            "pineapple",
            "banana",
            "jackfruit",
            "custard apple",
            "pomegranate",
            "hay",
            "carbonara",
            "chocolate syrup",
            "dough",
            "meatloaf",
            "pizza",
            "pot pie",
            "burrito",
            "red wine",
            "espresso",
            "cup",
            "eggnog",
            "alp",
            "bubble",
            "cliff",
            "coral reef",
            "geyser",
            "lakeshore",
            "promontory",
            "shoal",
            "seashore",
            "valley",
            "volcano",
            "baseball player",
            "bridegroom",
            "scuba diver",
            "rapeseed",
            "daisy",
            "yellow lady's slipper",
            "corn",
            "acorn",
            "rose hip",
            "horse chestnut seed",
            "coral fungus",
            "agaric",
            "gyromitra",
            "stinkhorn mushroom",
            "earth star",
            "hen-of-the-woods",
            "bolete",
            "ear",
            "toilet paper"]
